<?php

namespace Drupal\translators_content\Plugin;

use Drupal\Core\Access\AccessResult;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Language\LanguageInterface;
use Drupal\Core\Plugin\DefaultPluginManager;
use Drupal\content_translation\Access\ContentTranslationDeleteAccess;

/**
 * Class TranslatorsAccessRulesPluginManager.
 *
 * @package Drupal\translators_content\Access
 */
class TranslatorsAccessRulesPluginManager extends DefaultPluginManager {

  /**
   * Supported operations array.
   *
   * @var array
   */
  protected $supportedOperations;
  /**
   * Route match.
   *
   * @var \Drupal\Core\Routing\RouteMatchInterface
   */
  protected $routeMatch;
  /**
   * Current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $currentUser;
  /**
   * Entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;
  /**
   * Language manager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected $languageManager;
  /**
   * Config factory.
   *
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  protected $configFactory;
  /**
   * Config factory.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;
  /**
   * Translator skills service.
   *
   * @var \Drupal\translators\Services\TranslatorSkills
   */
  protected $translatorSkills;

  /**
   * The entity field manager.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * The content translation manager.
   *
   * @var \Drupal\content_translation\ContentTranslationManagerInterface
   */
  protected $contentTranslationManager;

  /**
   * Drupal core content_translation.delete_access service.
   *
   * @var \Drupal\content_translation\Access\ContentTranslationDeleteAccess
   */
  protected $defaultDeleteAccess;

  /**
   * {@inheritdoc}
   */
  public function __construct(\Traversable $namespaces, CacheBackendInterface $cache_backend, ModuleHandlerInterface $module_handler) {
    parent::__construct(
      'Plugin/TranslatorsAccessRules',
      $namespaces,
      $module_handler,
      'Drupal\translators_content\Plugin\TranslatorsAccessRulesInterface',
      'Drupal\translators_content\Annotation\TranslatorsAccessRule'
    );
    $this->supportedOperations = ['update', 'delete', 'create'];
    $this->alterInfo('translators_content_access_rules_info');
    $this->setCacheBackend($cache_backend, 'translators_content_access_rules_info_plugins');

    list($this->contentTranslationManager,
      $this->entityFieldManager,
      $this->translatorSkills,
      $this->languageManager,
      $this->entityTypeManager,
      $this->configFactory,
      $this->routeMatch,
      $this->currentUser) = array_reverse(func_get_args());
    $this->config = $this->configFactory->get('translators.settings');
    $this->defaultDeleteAccess = new ContentTranslationDeleteAccess($this->entityTypeManager, $this->contentTranslationManager);

  }

  /**
   * Check user access for the specified operation and language.
   *
   * @param string $operation
   *   Operation name.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   Entity to be operated on.
   * @param string|null $langcode
   *   Language ID.
   *
   * @return \Drupal\Core\Access\AccessResult|\Drupal\Core\Access\AccessResultAllowed|\Drupal\Core\Access\AccessResultForbidden|\Drupal\Core\Access\AccessResultNeutral
   *   Access result.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   * @throws \Drupal\Core\Entity\EntityMalformedException
   */
  public function checkAccess($operation, $entity = NULL, $langcode = NULL) {
    // Workaround for admins.
    if ($this->isAdmin()) {
      return AccessResult::allowed()
        ->cachePerPermissions();
    }

    // Skip any non-supported operations.
    if (!in_array($operation, $this->supportedOperations)) {
      return AccessResult::neutral()
        ->cachePerPermissions();
    }

    // If the permissions checking is disabled - re-use the core's flow.
    if (!$this->config->get('enable_translators_content_permissions')) {
      return $this->defaultCheckAccess($operation, $entity, $langcode);
    }

    $definitions = $this->getDefinitions();
    // If there are no rules or no entity - skip checking.
    if (empty($definitions) || !$entity instanceof ContentEntityInterface) {
      // Workaround for "create" operation.
      if ($operation === 'create') {
        return $this->processCreateOperation();
      }
      return AccessResult::neutral()
        ->cachePerPermissions();
    }

    // Workaround for user edit form access.
    if ($entity->getEntityTypeId() === 'user' && !$this->currentUser->isAnonymous()) {
      if ($entity->id() === $this->currentUser->id() && $operation === 'update') {
        // Just return the same code as in User module.
        // @see \Drupal\user\UserAccessControlHandler::checkAccess().
        return AccessResult::allowed()->cachePerUser();
      }
    }

    // Additional workaround for the "create translation" operation.
    if ($operation === 'create' && $entity instanceof ContentEntityInterface) {
      return $this->processCreateTranslationOperation($langcode);
    }

    // Try to find at least one rule that allowing user to access.
    foreach ($definitions as $id => $definition) {
      /** @var \Drupal\translators_content\Plugin\TranslatorsAccessRulesInterface $instance */
      $instance = $this->createInstance($id, $definition);
      if ($instance->isAllowed($operation, $entity, $langcode)) {
        return AccessResult::allowed()
          ->cachePerPermissions();
      }
    }

    return AccessResult::forbidden()
      ->cachePerPermissions();
  }

  /**
   * Process entity create translation operation access.
   *
   * @param string $target_langcode
   *   Language ID.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   Access result object.
   */
  protected function processCreateTranslationOperation($target_langcode) {
    if ($this->currentUser->hasPermission('create content translations')) {
      return AccessResult::allowed()->cachePerPermissions();
    }

    // Get target language.
    $translationTarget = $this->routeMatch->getParameter('target');

    if (!$translationTarget instanceof LanguageInterface) {
      $translationTarget = $this->languageManager->getLanguage($target_langcode);
    }

    // Fetch source language object from the route params.
    $translationSource = $this->routeMatch->getParameter('source');

    // Fallback for the non-specified source language.
    if (!$translationSource instanceof LanguageInterface) {
      // Get the users preferred language or the default translation language.
      $entity = $this->routeMatch->getParameter('node');
      if ($entity instanceof ContentEntityInterface) {
        $source_langcode = $this->translatorSkills->getTranslationSourceLangcode($entity, $translationTarget->getId());
        $translationSource = $this->languageManager->getLanguage($source_langcode);
      }
    }

    // Allow access if the user has both languages in their translation skills.
    if (!$this->config->get('enable_strict_translation_skill_pairing')) {
      return AccessResult::allowedIf(
        $translationSource instanceof LanguageInterface
        && $translationTarget instanceof LanguageInterface
        && $this->translatorSkills->hasLangcode($translationSource->getId())
        && $this->translatorSkills->hasLangcode($translationTarget->getId())
      )->cachePerPermissions();
    }

    // Allow access if the user has the source-target language pair.
    return AccessResult::allowedIf(
      $translationSource instanceof LanguageInterface
      && $translationTarget instanceof LanguageInterface
      && $this->translatorSkills->hasTranslationSkills($translationSource->getId(), $translationTarget->getId())
    )->cachePerPermissions();
  }

  /**
   * Process entity create operation access.
   *
   * @return \Drupal\Core\Access\AccessResult
   *   Access result object.
   */
  protected function processCreateOperation() {
    $langcode = $this->languageManager->getCurrentLanguage()->getId();
    foreach ($this->entityTypeManager->getDefinitions() as $entity_type) {
      if ($entity_type->getPermissionGranularity() === 'bundle') {
        $bundle_key = $entity_type->getBundleEntityType();
        if ($bundle = $this->routeMatch->getParameter($bundle_key)) {
          return AccessResult::allowedIfHasPermission(
            $this->currentUser,
            "translators_content create {$bundle->id()} content"
          )->andIf(
            AccessResult::allowedIf($this->translatorSkills->hasLangcode($langcode))
          )->cachePerPermissions();
        }
      }
    }
    return AccessResult::neutral()
      ->cachePerPermissions();
  }

  /**
   * Check whether the current user is admin or not.
   *
   * @return bool
   *   TRUE - if current user is admin, FALSE otherwise.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   * @throws \Drupal\Component\Plugin\Exception\PluginNotFoundException
   */
  protected function isAdmin() {
    // Prevent any checking for anonymous users.
    if ($this->currentUser->isAnonymous()) {
      return FALSE;
    }
    // Load all admin roles.
    $admin_roles = $this->entityTypeManager
      ->getStorage('user_role')
      ->getQuery()
      ->condition('is_admin', TRUE, '=')
      ->execute();

    // Get roles of current users.
    $roles = $this->currentUser->getRoles(TRUE);

    return !empty(array_intersect($roles, $admin_roles));
  }

  /**
   * Use Drupal core flow check access to translation.
   *
   * @param string $operation
   *   Operation name.
   * @param \Drupal\Core\Entity\ContentEntityInterface|null $entity
   *   Entity to be operated on.
   * @param string|null $langcode
   *   Language ID.
   *
   * @return \Drupal\Core\Access\AccessResult|\Drupal\Core\Access\AccessResultAllowed|\Drupal\Core\Access\AccessResultForbidden|\Drupal\Core\Access\AccessResultNeutral
   *   Access result.
   *
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  protected function defaultCheckAccess($operation, $entity = NULL, $langcode = NULL) {
    if (!$entity instanceof ContentEntityInterface) {
      return AccessResult::neutral();
    }

    $handler = $this->entityTypeManager->getHandler($entity->getEntityTypeId(), 'translation');
    $entity_type = $entity->getEntityType();
    $original = $entity->getUntranslated()->language()->getId();
    $is_original = $original == $langcode;

    if ($this->languageManager->isMultilingual()) {
      // Determine whether the current entity is translatable.
      $translatable = FALSE;
      foreach ($this->entityFieldManager->getFieldDefinitions($entity->getEntityTypeId(), $entity->bundle()) as $instance) {
        if ($instance->isTranslatable()) {
          $translatable = TRUE;
          break;
        }
      }

      $translations = $entity->getTranslationLanguages();
      switch ($operation) {
        case 'create':
          $source = $this->translatorSkills->getTranslationSourceLangcode($entity, $langcode);

          $create_translation_access = $handler->getTranslationAccess($entity, 'create');
          if ($source != $langcode && $create_translation_access->isAllowed()) {
            if ($translatable) {
              return AccessResult::allowed();
            }
          }
          break;

        case 'update':
          if (array_key_exists($langcode, $translations)) {
            $update_access = $entity->access('update', NULL, TRUE);
            $translation_access = $handler->getTranslationAccess($entity, 'update');
            if ($update_access->isAllowed() && $entity_type->hasLinkTemplate('edit-form')) {
              return AccessResult::allowed();
            }
            elseif (!$is_original && $translation_access->isAllowed()) {
              return AccessResult::allowed();
            }
          }
          break;

        case 'delete':
          if (array_key_exists($langcode, $translations)) {
            if (!$is_original) {
              $translation = $entity->getTranslation($langcode);
              $delete_route_access = $this->defaultDeleteAccess
                ->checkAccess($translation);
              if ($delete_route_access->isAllowed()) {
                $delete_access = $entity->access('delete', NULL, TRUE);
                $translation_access = $handler->getTranslationAccess($entity, 'delete');

                if ($delete_access->isAllowed() && $entity_type->hasLinkTemplate('delete-form')) {
                  return AccessResult::allowed();
                }
                elseif ($translation_access->isAllowed()) {
                  return AccessResult::allowed();
                }
              }
            }
          }
          break;
      }
    }

    return AccessResult::forbidden();
  }

}
