<?php

namespace Drupal\Tests\translators_content\Functional;

use Drupal\Component\Render\FormattableMarkup;
use Drupal\translators_content\TranslatorsContentTestsTrait;
use Drupal\Tests\BrowserTestBase;
use Drupal\Tests\node\Traits\NodeCreationTrait;
use Drupal\Core\Url;
use Drupal\user\UserInterface;

/**
 * Tests Drupal core permissions when module permissions aren't enabled.
 *
 * @package Drupal\Tests\translators_content\Functional
 *
 * @group translators_content
 */
class TranslatorsContentDefaultWorkflowsTest extends BrowserTestBase {
  use TranslatorsContentTestsTrait, NodeCreationTrait;

  /**
   * {@inheritdoc}
   */
  public $profile = 'standard';

  /**
   * {@inheritdoc}
   */
  public static $modules = ['translators_content'];

  /**
   * Node to test.
   *
   * @var object
   */
  protected $node;

  /**
   * The added languages.
   *
   * @var array
   */
  protected $langcodes;

  /**
   * {@inheritdoc}
   */
  public function setUp() {
    parent::setUp();
    $this->langcodes = $this->getAllTestingLanguages();
    // Make all required configurations before testing.
    $this->setUpTest();
  }

  /**
   * User has no permissions for translation except 'translate any entity'.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function testNonePermissions() {
    $user = $this->createUser(['translate any entity']);

    // Test workflows for the user with no permissions.
    $expected_status = [
      'edit' => 403,
      'delete' => 403,
      'overview' => 200,
      'add_translation' => 403,
      'edit_translation' => 403,
      'delete_translation' => 403,
    ];
    $this->doTestWorkflows($user, $expected_status);
  }

  /**
   * Test case for editor.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function testEditorPermissions() {
    $user = $this->createUser(['bypass node access']);

    // Test workflows for the editor.
    $expected_status = [
      'edit' => 200,
      'delete' => 200,
      'overview' => 403,
      'add_translation' => 403,
      'edit_translation' => 200,
      'delete_translation' => 200,
    ];
    $this->doTestWorkflows($user, $expected_status, TRUE);
  }

  /**
   * Test case for translator.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function testTranslatorPermissions() {
    $user = $this->createUser([
      'create content translations',
      'translate any entity',
      'update content translations',
      'delete content translations',
    ]);

    // Test workflows for the translator.
    $expected_status = [
      'edit' => 403,
      'delete' => 403,
      'overview' => 200,
      'add_translation' => 200,
      'edit_translation' => 200,
      'delete_translation' => 200,
    ];
    $this->doTestWorkflows($user, $expected_status);
  }

  /**
   * Test case for admin.
   *
   * @throws \Drupal\Core\Entity\EntityStorageException
   */
  public function testAdminPermissions() {
    $user = $this->createUser([
      'bypass node access',
      'create content translations',
      'translate any entity',
      'administer languages',
      'administer content translation',
    ]);

    // Test workflows for the admin.
    $expected_status = [
      'edit' => 200,
      'delete' => 200,
      'overview' => 200,
      'add_translation' => 200,
      'edit_translation' => 200,
      'delete_translation' => 200,
    ];
    $this->doTestWorkflows($user, $expected_status, TRUE);
  }

  /**
   * Additional steps for tests set up.
   */
  protected function setUpTest() {
    $this->drupalLogin($this->rootUser);
    $this->createLanguages();
    $this->enableTranslation('node', 'article');
    $this->enablePermissionsChecking(TRUE);
    $this->enableAllowAccessByStrictTranslationSkillsPairing(TRUE);
    $this->enableAutoPresetSourceLanguage();
    $this->enableFilterTranslationOverviewToSkills(TRUE);
    $this->createTestArticle();
    $this->drupalLogout();
  }

  /**
   * Create test article with default language.
   */
  protected function createTestArticle() {
    $node = $this
      ->drupalCreateNode([
        'status' => TRUE,
        'type' => 'article',
        'langcode' => $this->defaultLanguage,
      ]);
    $node->save();

    $this->node = $node;

    // Create a translation.
    $add_translation_url = Url::fromRoute(
      "entity.node.content_translation_add",
      [
        'node' => $this->node->id(),
        'source' => $this->defaultLanguage,
        'target' => $this->langcodes[3],
      ]
    );
    $this->drupalPostForm($add_translation_url, [], t('Save'));
    $this->rebuildContainer();
  }

  /**
   * Checks that workflows have the expected behaviors for the given user.
   *
   * @param \Drupal\user\UserInterface $user
   *   The user to test the workflow behavior against.
   * @param array $expected_status
   *   The an associative array with the operation name as key and the expected
   *   status as value.
   * @param bool $skip_link_check
   *   We want to skip link check for links if user has no access to translation
   *   overview, but has access to node actions.
   */
  protected function doTestWorkflows(UserInterface $user, array $expected_status, $skip_link_check = FALSE) {
    $default_langcode = $this->defaultLanguage;
    $languages = $this->container->get('language_manager')->getLanguages();
    $args = ['@user_label' => $user->getAccountName()];
    $options = ['language' => $languages[$default_langcode], 'absolute' => TRUE];
    $this->drupalLogin($user);

    // Check whether the user is allowed to access the entity form in edit mode.
    $edit_url = $this->node->toUrl('edit-form', $options);
    $this->drupalGet($edit_url, $options);
    $this->assertResponse($expected_status['edit'], new FormattableMarkup('The @user_label has the expected edit access.', $args));

    // Check whether the user is allowed to access the entity delete form.
    $delete_url = $this->node->toUrl('delete-form', $options);
    $this->drupalGet($delete_url, $options);
    $this->assertResponse($expected_status['delete'], new FormattableMarkup('The @user_label has the expected delete access.', $args));

    // Check whether the user is allowed to access the translation overview.
    $langcode = $this->langcodes[1];
    $options['language'] = $languages[$langcode];
    $translations_url = $this->node->toUrl('drupal:content-translation-overview', $options)->toString();
    $this->drupalGet($translations_url);
    $this->assertResponse($expected_status['overview'], new FormattableMarkup('The @user_label has the expected translation overview access.', $args));

    // Check whether the user is allowed to create a translation.
    $add_translation_url = Url::fromRoute(
      "entity.node.content_translation_add",
      [
        'node' => $this->node->id(),
        'source' => $default_langcode,
        'target' => $langcode,
      ],
      $options
    );
    if ($expected_status['add_translation'] == 200) {
      $this->clickLink('Add');
      $this->assertUrl($add_translation_url->toString(), [], 'The translation overview points to the translation form when creating translations.');
    }
    else {
      $this->drupalGet($add_translation_url);
    }
    $this->assertResponse($expected_status['add_translation'], new FormattableMarkup('The @user_label has the expected translation creation access.', $args));

    // Check whether the user is allowed to edit a translation.
    $langcode = $this->langcodes[3];
    $options['language'] = $languages[$langcode];
    $edit_translation_url = Url::fromRoute(
      "entity.node.content_translation_edit",
      [
        'node' => $this->node->id(),
        'language' => $langcode,
      ],
      $options
    );
    if ($expected_status['edit_translation'] == 200 && !$skip_link_check) {
      $this->drupalGet($translations_url);
      $editor = $expected_status['edit'] == 200;

      if ($editor) {
        $this->clickLink('Edit', 2);
        // An editor should be pointed to the entity form in multilingual mode.
        // We need a new expected edit path with a new language.
        $expected_edit_path = $this->node->toUrl('edit-form', $options)->toString();
        $this->assertUrl($expected_edit_path, [], 'The translation overview points to the edit form for editors when editing translations.');
      }
      else {
        $this->clickLink('Edit');
        // While a translator should be pointed to the translation form.
        $this->assertUrl($edit_translation_url->toString(), [], 'The translation overview points to the translation form for translators when editing translations.');
      }
    }
    else {
      $this->drupalGet($edit_translation_url);
    }
    $this->assertResponse($expected_status['edit_translation'], new FormattableMarkup('The @user_label has the expected translation edit access.', $args));

    // Check whether the user is allowed to delete a translation.
    $langcode = $this->langcodes[3];
    $options['language'] = $languages[$langcode];
    $delete_translation_url = Url::fromRoute(
      "entity.node.content_translation_delete",
      [
        'node' => $this->node->id(),
        'language' => $langcode,
      ],
      $options
    );
    if ($expected_status['delete_translation'] == 200 && !$skip_link_check) {
      $this->drupalGet($translations_url);
      $editor = $expected_status['delete'] == 200;

      if ($editor) {
        $this->clickLink('Delete', 2);
        // An editor should be pointed to the entity deletion form in
        // multilingual mode. We need a new expected delete path with a new
        // language.
        $expected_delete_path = $this->node->toUrl('delete-form', $options)->toString();
        $this->assertUrl($expected_delete_path, [], 'The translation overview points to the delete form for editors when deleting translations.');
      }
      else {
        $this->clickLink('Delete');
        // While a translator should be pointed to the translation deletion
        // form.
        $this->assertUrl($delete_translation_url->toString(), [], 'The translation overview points to the translation deletion form for translators when deleting translations.');
      }
    }
    else {
      $this->drupalGet($delete_translation_url);
    }
    $this->assertResponse($expected_status['delete_translation'], new FormattableMarkup('The @user_label has the expected translation deletion access.', $args));
  }

}
